﻿#nullable enable

using System;
using System.Collections.Generic;
using Microsoft.Extensions.Configuration;
using Microsoft.Extensions.DependencyInjection;
using Microsoft.Extensions.Hosting;
using Microsoft.Extensions.Logging;
using Unity.Extensions.Configuration;
using Unity.Extensions.Logging;

namespace Unity.Extensions.GenericHost
{
    public static class GameHostBuilderExtensions
    {
        public static IHostBuilder UseEnvironment(this IHostBuilder hostBuilder, string environment)
        {
            return hostBuilder.ConfigureHostConfiguration(configBuilder =>
            {
                if (string.IsNullOrEmpty(environment))
                {
                    throw new ArgumentNullException(nameof(environment));
                }

                configBuilder.AddInMemoryCollection(new[]
                {
                    new KeyValuePair<string, string?>(HostDefaults.EnvironmentKey, environment)
                });
            });
        }
        
        public static IHostBuilder UseContentRoot(this IHostBuilder hostBuilder, string contentRoot)
        {
            return hostBuilder.ConfigureHostConfiguration(configBuilder =>
            {
                if (string.IsNullOrEmpty(contentRoot))
                {
                    throw new ArgumentNullException(nameof(contentRoot));
                }

                configBuilder.AddInMemoryCollection(new[]
                {
                    new KeyValuePair<string, string?>(HostDefaults.ContentRootKey, contentRoot)
                });
            });
        }
        
        public static IHostBuilder UseDefaultServiceProvider(this IHostBuilder hostBuilder, Action<ServiceProviderOptions> configure)
            => hostBuilder.UseDefaultServiceProvider((_, options) => configure(options));
        
        public static IHostBuilder UseDefaultServiceProvider(this IHostBuilder hostBuilder, Action<HostBuilderContext, ServiceProviderOptions> configure)
        {
            return hostBuilder.UseServiceProviderFactory(context =>
            {
                var options = new ServiceProviderOptions();
                configure(context, options);
                return new DefaultServiceProviderFactory(options);
            });
        }
        
        public static IHostBuilder ConfigureLogging(this IHostBuilder hostBuilder, Action<HostBuilderContext, ILoggingBuilder> configureLogging)
        {
            return hostBuilder.ConfigureServices((context, collection) => collection.AddLogging(builder => configureLogging(context, builder)));
        }
        
        public static IHostBuilder ConfigureLogging(this IHostBuilder hostBuilder, Action<ILoggingBuilder> configureLogging)
        {
            return hostBuilder.ConfigureServices((_, collection) => collection.AddLogging(configureLogging));
        }
        
        public static IHostBuilder ConfigureHostOptions(this IHostBuilder hostBuilder, Action<HostBuilderContext, HostOptions> configureOptions)
        {
            return hostBuilder.ConfigureServices(
                (context, collection) => collection.Configure<HostOptions>(options => configureOptions(context, options)));
        }
        
        public static IHostBuilder ConfigureHostOptions(this IHostBuilder hostBuilder, Action<HostOptions> configureOptions)
        {
            return hostBuilder.ConfigureServices(collection => collection.Configure(configureOptions));
        }
        
        public static IHostBuilder ConfigureAppConfiguration(this IHostBuilder hostBuilder, Action<IConfigurationBuilder> configureDelegate)
        {
            return hostBuilder.ConfigureAppConfiguration((_, builder) => configureDelegate(builder));
        }
        
        public static IHostBuilder ConfigureServices(this IHostBuilder hostBuilder, Action<IServiceCollection> configureDelegate)
        {
            return hostBuilder.ConfigureServices((_, collection) => configureDelegate(collection));
        }
        
        public static IHostBuilder ConfigureContainer<TContainerBuilder>(this IHostBuilder hostBuilder, Action<TContainerBuilder> configureDelegate)
        {
            return hostBuilder.ConfigureContainer<TContainerBuilder>((_, builder) => configureDelegate(builder));
        }
        
        public static IHostBuilder ConfigureDefaults(this IHostBuilder builder)
        {
            return builder.ConfigureHostConfiguration(ApplyDefaultHostConfiguration)
                .ConfigureAppConfiguration(ApplyDefaultAppConfiguration)
                .ConfigureServices(AddDefaultServices)
                .UseServiceProviderFactory(context => new DefaultServiceProviderFactory(CreateDefaultServiceProviderOptions(context)));
        }
        
        private static void ApplyDefaultHostConfiguration(IConfigurationBuilder hostConfigBuilder)
        {
            SetDefaultContentRoot(hostConfigBuilder);

            hostConfigBuilder.AddEnvironmentVariables(prefix: "UNITY_");
        }
        
        internal static void SetDefaultContentRoot(IConfigurationBuilder hostConfigBuilder)
        {
            hostConfigBuilder.AddInMemoryCollection(new[]
            {
                new KeyValuePair<string, string?>(HostDefaults.ContentRootKey, UnityEngine.Application.persistentDataPath),
            });
        }
        
        internal static void ApplyDefaultAppConfiguration(HostBuilderContext hostingContext, IConfigurationBuilder appConfigBuilder)
        {
            var env = hostingContext.HostingEnvironment;
            
            appConfigBuilder.AddUnityResourcesConfiguration("appsettings.json", optional: true)
                .AddUnityResourcesConfiguration($"appsettings.{env.EnvironmentName}.json", optional: true);

            appConfigBuilder.AddEnvironmentVariables();
        }
        
        internal static void AddDefaultServices(HostBuilderContext hostingContext, IServiceCollection services)
        {
            services.AddLogging(logging =>
            {
                logging.AddConfiguration(hostingContext.Configuration.GetSection("Logging"));
                logging.AddUnityLogger();
            });
        }
        
        internal static ServiceProviderOptions CreateDefaultServiceProviderOptions(HostBuilderContext context)
        {
            var isDevelopment = context.HostingEnvironment.IsDevelopment();
            return new ServiceProviderOptions
            {
                ValidateScopes = isDevelopment,
                ValidateOnBuild = isDevelopment,
            };
        }
    }
}