﻿#nullable enable

using System;
using System.Collections.Generic;
using System.Linq;
using System.Runtime.ExceptionServices;
using System.Threading;
using System.Threading.Tasks;
using Microsoft.Extensions.DependencyInjection;
using Microsoft.Extensions.FileProviders;
using Microsoft.Extensions.Hosting;
using Microsoft.Extensions.Logging;
using Microsoft.Extensions.Options;

namespace Unity.Extensions.GenericHost.Internal
{
    internal sealed class GameHost : IHost, IAsyncDisposable
    {
        private readonly ILogger<GameHost> _logger;
        private readonly IHostLifetime _hostLifetime;
        private readonly UnityApplicationLifetime _applicationLifetime;
        private readonly HostOptions _options;
        private readonly IHostEnvironment _hostEnvironment;
        private readonly PhysicalFileProvider _defaultProvider;
        private IEnumerable<IHostedService>? _hostedServices;
        private IEnumerable<IHostedLifecycleService>? _hostedLifecycleServices;
        private bool _hostStarting;
        private bool _hostStopped;
        
        private bool IsRunning => _applicationLifetime.ApplicationStarted.IsCancellationRequested && !_hostStopped;
        
        public IServiceProvider Services { get; }

        public GameHost(IServiceProvider services,
            IHostEnvironment hostEnvironment,
            PhysicalFileProvider defaultProvider,
            IHostApplicationLifetime applicationLifetime,
            ILogger<GameHost> logger,
            IHostLifetime hostLifetime,
            IOptions<HostOptions> options)
        {
            Services = services;
            _applicationLifetime = (applicationLifetime as UnityApplicationLifetime)!;
            _hostEnvironment = hostEnvironment;
            _defaultProvider = defaultProvider;

            if (_applicationLifetime is null)
            {
                throw new ArgumentException();
            }
            _logger = logger;
            _hostLifetime = hostLifetime;
            _options = options?.Value ?? throw new ArgumentNullException(nameof(options));
        }

        public async Task StartAsync(CancellationToken cancellationToken = default)
        {
            _logger.LogDebug(eventId: LoggerEventIds.Starting, message: "Hosting starting");

            using var cts = CancellationTokenSource.CreateLinkedTokenSource(cancellationToken, _applicationLifetime.ApplicationStopping);
            if (_options.StartupTimeout != Timeout.InfiniteTimeSpan)
            {
                cts.CancelAfter(_options.StartupTimeout);
            }
                
            cancellationToken = cts.Token;
                
            await _hostLifetime.WaitForStartAsync(cancellationToken).ConfigureAwait(false);
            cancellationToken.ThrowIfCancellationRequested();
                
            List<Exception> exceptions = new();
            _hostedServices ??= Services.GetRequiredService<IEnumerable<IHostedService>>();
            _hostedLifecycleServices = GetHostLifecycles(_hostedServices);
            _hostStarting = true;
            var concurrent = _options.ServicesStartConcurrently;
            var abortOnFirstException = !concurrent;
                
            var validator = Services.GetService<IStartupValidator>();
            if (validator is not null)
            {
                try
                {
                    validator.Validate();
                }
                catch (Exception e)
                {
                    exceptions.Add(e);
                    LogAndRethrow();
                }
            }

            if (_hostedLifecycleServices is not null)
            {
                await ForeachService(_hostedLifecycleServices,
                        cancellationToken,
                        concurrent,
                        abortOnFirstException,
                        exceptions,
                        (service,
                            token) => service.StartingAsync(token))
                    .ConfigureAwait(false);
                    
                LogAndRethrow();
            }

            await ForeachService(_hostedServices, cancellationToken, concurrent, abortOnFirstException, exceptions,
                async (service, token) =>
                {
                    await service.StartAsync(token).ConfigureAwait(false);

                    if (service is BackgroundService backgroundService)
                    {
                        _ = TryExecuteBackgroundServiceAsync(backgroundService);
                    }
                }).ConfigureAwait(false);
                
            LogAndRethrow();
                
            if (_hostedLifecycleServices is not null)
            {
                await ForeachService(_hostedLifecycleServices, cancellationToken, concurrent, abortOnFirstException, exceptions,
                    (service, token) => service.StartedAsync(token)).ConfigureAwait(false);
            }
                
            LogAndRethrow();
                
            _applicationLifetime.NotifyStarted();
            return;

            void LogAndRethrow()
            {
                switch (exceptions.Count)
                {
                    case <= 0:
                    {
                        return;
                    }
                    case 1:
                    {
                        // Rethrow if it's a single error
                        var singleException = exceptions[0];
                        _logger.LogError(
                            eventId: LoggerEventIds.HostedServiceStartupFaulted,
                            exception: singleException,
                            message: "Hosting failed to start");
                        ExceptionDispatchInfo.Capture(singleException).Throw();
                        break;
                    }
                    default:
                    {
                        var ex = new AggregateException("One or more hosted services failed to start.", exceptions);
                        _logger.LogError(
                            eventId: LoggerEventIds.HostedServiceStartupFaulted,
                            exception: ex,
                            message: "Hosting failed to start");
                        throw ex;
                    }
                }
            }
        }

        public async Task StopAsync(CancellationToken cancellationToken = default)
        {
            _logger.LogDebug(eventId: LoggerEventIds.Stopping, message: "Hosting stopping");
            
            CancellationTokenSource? cts = null;
            if (_options.ShutdownTimeout != Timeout.InfiniteTimeSpan)
            {
                cts = CancellationTokenSource.CreateLinkedTokenSource(cancellationToken);
                cts.CancelAfter(_options.ShutdownTimeout);
                cancellationToken = cts.Token;
            }

            using (cts)
            {
                List<Exception> exceptions = new();

                if (!_hostStarting)
                {
                    _applicationLifetime.StopApplication();
                }
                else
                {
                    UnityEngine.Debug.Assert(_hostedServices != null);
                    
                    var reversedServices = _hostedServices!.Reverse();
                    var reversedLifetimeServices = _hostedLifecycleServices?.Reverse();
                    var concurrent = _options.ServicesStopConcurrently;
                    
                    if (reversedLifetimeServices is not null)
                    {
                        await ForeachService(reversedLifetimeServices, cancellationToken, concurrent, abortOnFirstException: false, exceptions,
                            (service, token) => service.StoppingAsync(token)).ConfigureAwait(false);
                    }
                    
                    _applicationLifetime.StopApplication();
                    
                    await ForeachService(reversedServices, cancellationToken, concurrent, abortOnFirstException: false, exceptions, (service, token) =>
                        service.StopAsync(token)).ConfigureAwait(false);
                    
                    if (reversedLifetimeServices is not null)
                    {
                        await ForeachService(reversedLifetimeServices, cancellationToken, concurrent, abortOnFirstException: false, exceptions, (service, token) =>
                            service.StoppedAsync(token)).ConfigureAwait(false);
                    }
                }
                
                _applicationLifetime.NotifyStopped();

                try
                {
                    await _hostLifetime.StopAsync(cancellationToken).ConfigureAwait(false);
                }
                catch (Exception e)
                {
                    exceptions.Add(e);
                }
                
                _hostStopped = true;
                
                if (exceptions.Count > 0)
                {
                    if (exceptions.Count == 1)
                    {
                        // Rethrow if it's a single error
                        var singleException = exceptions[0];
                        _logger.LogDebug(
                            eventId: LoggerEventIds.StoppedWithException,
                            exception: singleException,
                            message: "Hosting shutdown exception");
                        ExceptionDispatchInfo.Capture(singleException).Throw();
                    }
                    else
                    {
                        var ex = new AggregateException("One or more hosted services failed to stop.", exceptions);
                        _logger.LogDebug(
                            eventId: LoggerEventIds.StoppedWithException,
                            exception: ex,
                            message: "Hosting shutdown exception");
                        throw ex;
                    }
                }
            }
        }
        
        public void Dispose() => DisposeAsync().AsTask().GetAwaiter().GetResult();
        
        public async ValueTask DisposeAsync()
        {
            var contentRootFileProvider = _hostEnvironment.ContentRootFileProvider;
            await DisposeAsync(contentRootFileProvider).ConfigureAwait(false);

            if (!ReferenceEquals(contentRootFileProvider, _defaultProvider))
            {
                // In the rare case that the user replaced the ContentRootFileProvider, dispose it and the one
                // we originally created
                await DisposeAsync(_defaultProvider).ConfigureAwait(false);
            }

            // Dispose the service provider
            await DisposeAsync(Services).ConfigureAwait(false);

            static ValueTask DisposeAsync(object o)
            {
                switch (o)
                {
                    case IAsyncDisposable asyncDisposable:
                        return asyncDisposable.DisposeAsync();
                    case IDisposable disposable:
                        disposable.Dispose();
                        break;
                }
                return default;
            }
        }

        private async Task TryExecuteBackgroundServiceAsync(BackgroundService backgroundService)
        {
            var backgroundTask = backgroundService.ExecuteTask;
            if (backgroundTask is null)
            {
                return;
            }

            try
            {
                await backgroundTask.ConfigureAwait(false);
            }
            catch (Exception e)
            {
                if (_applicationLifetime.ApplicationStopping.IsCancellationRequested && backgroundTask.IsCanceled && e is OperationCanceledException)
                {
                    return;
                }
                
                _logger.LogCritical(
                    eventId: LoggerEventIds.BackgroundServiceStoppingHost,
                    exception: e,
                    message: "A BackgroundService has thrown an unhandled exception, and the IHost instance is stopping");

                _applicationLifetime.StopApplication();
            }
        }

        private static async Task ForeachService<T>(
            IEnumerable<T> services,
            CancellationToken token,
            bool concurrent,
            bool abortOnFirstException,
            List<Exception> exceptions,
            Func<T, CancellationToken, Task> operation)
        {
            if (concurrent)
            {
                // The beginning synchronous portions of the implementations are run serially in registration order for
                // performance since it is common to return Task.Completed as a noop.
                // Any subsequent asynchronous portions are grouped together and run concurrently.
                List<Task>? tasks = null;

                foreach (var service in services)
                {
                    Task task;
                    try
                    {
                        task = operation(service, token);
                    }
                    catch (Exception ex)
                    {
                        exceptions.Add(ex); // Log exception from sync method.
                        continue;
                    }

                    if (task.IsCompleted)
                    {
                        if (task.Exception is not null)
                        {
                            exceptions.AddRange(task.Exception.InnerExceptions); // Log exception from async method.
                        }
                    }
                    else
                    {
                        // The task encountered an await; add it to a list to run concurrently.
                        tasks ??= new List<Task>();
                        tasks.Add(task);
                    }
                }

                if (tasks is not null)
                {
                    var groupedTasks = Task.WhenAll(tasks);

                    try
                    {
                        await groupedTasks.ConfigureAwait(false);
                    }
                    catch (Exception ex)
                    {
                        if (groupedTasks.IsFaulted)
                        {
                            exceptions.AddRange(groupedTasks.Exception.InnerExceptions);
                        }
                        else
                        {
                            exceptions.Add(ex);
                        }
                    }
                }
            }
            else
            {
                foreach (var service in services)
                {
                    try
                    {
                        await operation(service, token).ConfigureAwait(false);
                    }
                    catch (Exception ex)
                    {
                        exceptions.Add(ex);
                        if (abortOnFirstException)
                        {
                            return;
                        }
                    }
                }
            }
        }

        private static List<IHostedLifecycleService>? GetHostLifecycles(IEnumerable<IHostedService> hostedServices)
        {
            List<IHostedLifecycleService>? result = null;

            foreach (var hostedService in hostedServices)
            {
                if (hostedService is IHostedLifecycleService service)
                {
                    result ??= new List<IHostedLifecycleService>();
                    result.Add(service);
                }
            }

            return result;
        }
    }
}