﻿#nullable enable

using System;
using System.Threading;
using System.Threading.Tasks;
using Microsoft.Extensions.Hosting;
using Microsoft.Extensions.Logging;

namespace Unity.Extensions.GenericHost.Internal
{
    internal class UnityLifetime : IHostLifetime, IDisposable
    {
        private CancellationTokenRegistration _applicationStartedRegistration;
        private CancellationTokenRegistration _applicationStoppingRegistration;
        private CancellationTokenRegistration _applicationStoppedRegistration;
        private IHostEnvironment Environment { get; }
        private IHostApplicationLifetime ApplicationLifetime { get; }
        private ILogger Logger { get; }

        public UnityLifetime(IHostEnvironment environment, IHostApplicationLifetime applicationLifetime, ILoggerFactory loggerFactory)
        {
            Environment = environment;
            ApplicationLifetime = applicationLifetime;
            Logger = loggerFactory.CreateLogger("Unity.Hosting.Lifetime");
        }

        public Task WaitForStartAsync(CancellationToken cancellationToken)
        {
            _applicationStartedRegistration = ApplicationLifetime.ApplicationStarted.Register(state =>
            {
                if (state is UnityLifetime unityLifetime)
                {
                    unityLifetime.OnApplicationStarted();
                }
            }, this);

            _applicationStoppingRegistration = ApplicationLifetime.ApplicationStopping.Register(state =>
            {
                if (state is UnityLifetime unityLifetime)
                {
                    unityLifetime.OnApplicationStopping();
                }
            }, this);
            
            _applicationStoppedRegistration = ApplicationLifetime.ApplicationStopped.Register(state =>
            {
                if (state is UnityLifetime unityLifetime)
                {
                    unityLifetime.OnApplicationStopped();
                }
            }, this);

            RegisterShutdownHandlers();
            
            return Task.CompletedTask;
        }

        public Task StopAsync(CancellationToken cancellationToken) => Task.CompletedTask; // NOP

        private void RegisterShutdownHandlers()
        {
            UnityEngine.Application.quitting += ProcessApplicationQuitting;
        }

        private void UnregisterShutdownHandlers()
        {
            UnityEngine.Application.quitting -= ProcessApplicationQuitting;
        }

        private void ProcessApplicationQuitting()
        {
            ApplicationLifetime.StopApplication();
        }

        private void OnApplicationStarted()
        {
            Logger.LogInformation("Application started");
            Logger.LogInformation("Hosting environment: {EnvName}", Environment.EnvironmentName);
            Logger.LogInformation("Content root path: {ContentRoot}", Environment.ContentRootPath);
        }
        
        private void OnApplicationStopping()
        {
            Logger.LogInformation("Application is shutting down...");
        }
        
        private void OnApplicationStopped()
        {
            Logger.LogInformation("Application stopped");
        }

        public void Dispose()
        {
            UnregisterShutdownHandlers();
            
            _applicationStartedRegistration.Dispose();
            _applicationStoppingRegistration.Dispose();
            _applicationStoppedRegistration.Dispose();
        }
    }
}